from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from clusterenv._types import Jobs, JobStatus, Nodes
from typing import Protocol
import logging
import gymnasium as gym


class NodeGenerator(Protocol):
    @classmethod
    def generate(cls, n: int, r: int, t: int, max_usage: float) -> Nodes:
        nodes: npt.NDArray[float] = np.full(
            (n, r, t),
            fill_value=max_usage,
            dtype=float,
        )
        return Nodes(nodes)


class JobGenerator(Protocol):
    LENGTH = dict(option=[0.0, 0.2, 0.3], prob=[0.7, 0.2, 0.1])
    USAGE = dict(option=[1], prob=[1])
    ARRIVAL_TIME = dict(option=[0.0], prob=[1])

    @classmethod
    def generate(cls, n: int, r: int, t: int, max_usage: float) -> Jobs:
        submission: np.array = t * np.random.choice(
            cls.ARRIVAL_TIME["option"], size=n, p=cls.ARRIVAL_TIME["prob"]
        ).astype(int)
        length: np.array = 1 + t * np.random.choice(
            cls.LENGTH["option"], size=(n, r), p=cls.LENGTH["prob"]
        )
        usage: np.array = max_usage * np.random.choice(
            cls.USAGE["option"], size=n, p=cls.USAGE["prob"]
        ).astype(float)
        usage: np.ndarray = np.tile(usage[..., np.newaxis, np.newaxis], (r, t))
        mask = np.arange(usage.shape[-1]) >= length[..., np.newaxis]
        usage[mask] = 0.0
        return Jobs(usage=usage, submission=submission)


@dataclass(slots=True)
class Cluster:
    """
    Attributes
    ------
        nodes (Nodes): TODO
        jobs (Jobs): TODO

    """

    nodes: Nodes
    jobs: Jobs
    _logger: logging.Logger = field(init=False)
    _run_time: npt.NDArray[int] = field(init=False)
    _max_val: float = field(init=False)
    _time: int = 0

    @property
    def observation(self) -> dict:
        return dict(
            Usage=self.nodes.usage,
            Queue=self.jobs.usage,
            Nodes=self.nodes.original,
            Status=self.jobs.status,
        )

    @property
    def time(self) -> int:
        return self._time

    @property
    def max_usage(self) -> float:
        return self._max_val

    def action_space(self) -> gym.Space:
        return gym.spaces.Discrete(len(self.nodes) * len(self.jobs) + 1)

    def observation_shape(self) -> gym.Space:
        max_val: float = np.max(self.nodes.original)
        return gym.spaces.Dict(
            dict(
                Usage=gym.spaces.Box(
                    low=0, high=max_val, shape=self.jobs.usage.shape, dtype=float
                ),
                Queue=gym.spaces.Box(
                    low=-1,
                    high=max_val,
                    shape=self.jobs.usage.shape,
                    dtype=float,
                ),
                Nodes=gym.spaces.Box(
                    low=0, high=max_val, shape=self.nodes.original.shape, dtype=float
                ),
                Status=gym.spaces.Box(
                    low=0, high=5, shape=self.jobs.status.shape, dtype=float
                ),
            )
        )

    def has_terminate(self) -> bool:
        return len(self.jobs.index_by_status(JobStatus.COMPLETE)) == len(self.jobs)

    def has_additional_jobs(self) -> bool:
        n_pending: int = len(self.jobs.index_by_status(JobStatus.COMPLETE))
        n_complete: int = len(self.jobs.index_by_status(JobStatus.RUNNING))
        return n_pending + n_complete != len(self.jobs)

    def total_wait_time(self):
        return self.jobs.total_pending_time()

    def __post_init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._run_time = np.zeros(len(self.jobs)).astype(float)
        self._max_val = np.max(self.nodes)
        self.jobs.check_arrived_jobs(self._time)

    def tick(self) -> None:
        """Forward cluster time by tick.
        Updating running time of all running jobs by one, show new income jobs.
        """
        self._logger.info(f"Tick {self._time}->{self._time+1}")
        self.jobs.inc(JobStatus.RUNNING)
        self._time += 1
        complete_jobs_idx: npt.ArrayLike[int] = self.jobs.check_complete_jobs()
        arrived_jobs_idx: npt.ArrayLike[int] = self.jobs.check_arrived_jobs(self._time)
        self._logger.info(f"Receive new jobs: {arrived_jobs_idx}")
        self._logger.info(f"Complete jobs: {complete_jobs_idx}")
        self.nodes.tick()

    def schedule(self, n_idx: int, j_idx: int) -> bool:
        job_status, _, job_usage = self.jobs[j_idx]
        match job_status:
            case JobStatus.PENDING:
                free_n_space: np.ndarray = self.nodes.free_space(n_idx)
                job_can_be_schedule: bool = bool(np.all(free_n_space >= job_usage))
                if job_can_be_schedule:
                    self.nodes[n_idx] += job_usage
                    self.jobs[j_idx] = self.jobs[j_idx].change_status(JobStatus.RUNNING)
                    logging.info(f"Succeed Allocated j.{j_idx} to n.{n_idx}")
                else:
                    logging.info(f"Can't Allocate j.{j_idx} n.{n_idx}, not enough resource")
                return job_can_be_schedule
            case _:
                logging.info(f"Can't Allocate j.{j_idx} with status {JobStatus(job_status).name}")
                return False



from clusterenv._types import JobStatus
from typing import Iterable, Callable, Any
from dataclasses import dataclass, field
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import matplotlib.pyplot as plt
import numpy.typing as npt
import numpy as np
import math


@dataclass
class ClusterRenderer:
    """
    Craete Gym Cluster Object.
    Allow To Represent Cluster logic: Scheduling, TimeTick, Generating Job.
    Job & Node reperesentation should be similar.

    :param nodes: List of nodes repersentation
    :param jobs: List of job representation
    :param resource: Number of Job/Node resource type
    :param max_time: Max Job Run time
    :param cooldown: Render cooldown between stepes
    """

    nodes: int
    jobs: int
    resource: int
    time: int
    cooldown: float
    fig: Figure = field(init=False)
    axs: npt.NDArray = field(init=False)
    CMAP_COLORS: tuple = ("copper", "gray", "twilight", "summer")
    REGULAR_COLOR: str = "copper"
    REGULAR_TITLE_COLOR: str = "black"
    ERROR_COLOR: str = "RdGy"
    CORRECT_COLOR: str = "Greens"
    ERROR_TITLE_COLOR: str = "red"

    def __post_init__(self):
        self.jobs_n_columns: int = math.ceil(self.jobs**0.5)
        self.nodes_n_columns: int = math.ceil(self.nodes**0.5)

        jobs_n_rows: int = math.ceil(self.jobs / self.jobs_n_columns)
        nodes_n_rows: int = math.ceil(self.nodes / self.nodes_n_columns)

        n_rows: int = max(jobs_n_rows, nodes_n_rows)
        n_columns: int = self.nodes_n_columns + self.jobs_n_columns

        self.fig, self.axs = plt.subplots(
            n_rows, n_columns, figsize=(8, 4), facecolor="white"
        )
        self._hide_unused(
            self.axs,
            nodes=self.nodes,
            jobs=self.jobs,
            nodes_n_columns=self.nodes_n_columns,
        )

    @classmethod
    def _hide_unused(cls, axs: np.ndarray, nodes: int, jobs: int, nodes_n_columns: int):
        nodes_to_remove: Iterable[Axes] = axs[:, :nodes_n_columns].flatten()[nodes:]
        jobs_to_remove: Iterable[Axes] = axs[:, nodes_n_columns:].flatten()[jobs:]
        for ax in nodes_to_remove:
            plt.delaxes(ax)
        for ax in jobs_to_remove:
            plt.delaxes(ax)

    @classmethod
    def _draw(
        cls,
        matrix: np.ndarray,
        /,
        *,
        title: str,
        title_color: str,
        ax: Axes,
        time: int,
        resource: int,
        cmap: str,
    ):
        ax.imshow(matrix, cmap=cmap, vmin=0, vmax=100)
        ax.set_title(title, fontsize=10, color=title_color)
        ax.set_xticks(np.arange(0, time, 0.5), minor=True)
        ax.set_yticks(np.arange(0, resource, 0.5), minor=True)
        ax.tick_params(which="minor", length=0)
        ax.grid(which="both", color="black", linestyle="-", linewidth=0.5, alpha=0.3)
        ax.set_xticks([])
        ax.set_yticks([])

    @classmethod
    def _draw_job(
        cls,
        job: np.ndarray,
        /,
        *,
        title_color: str,
        idx: int,
        ax: Axes,
        time: int,
        resource: int,
        cmap: str,
        status: JobStatus,
    ):
        title: str = f"[J.{idx}]"
        cmap = cmap if cmap == cls.ERROR_COLOR else cls.CMAP_COLORS[status]
        cls._draw(
            job,
            title=title,
            title_color=title_color,
            ax=ax,
            time=time,
            resource=resource,
            cmap=cmap,
        )

    @classmethod
    def _draw_node(
        cls,
        node: np.ndarray,
        /,
        *,
        title_color: str,
        idx: int,
        ax: Axes,
        time: int,
        resource: int,
        cmap: str,
    ):
        title: str = f"[N.{idx}]"
        cls._draw(
            node,
            title=title,
            title_color=title_color,
            ax=ax,
            time=time,
            resource=resource,
            cmap=cmap,
        )

    @classmethod
    def color_map(cls, error, correct):
        def inner(idx, pos):
            if error and idx == error[pos]:
                return cls.ERROR_COLOR
            elif correct and idx == correct[pos]:
                return cls.CORRECT_COLOR
            return cls.REGULAR_COLOR
        return inner

    def __call__(
        self,
        obs: dict[str, np.ndarray],
        /,
        *,
        current_time: int,
        error: None | tuple[int, int],
        correct: None | tuple[int, int]
    ) -> Any:
        self.fig.suptitle(f"Time: {current_time}", fontsize=16, fontweight="bold")
        nodes: npt.NDArray = obs["Usage"]
        queue: npt.NDArray = obs["Queue"]
        status: npt.NDArray[np.uint32] = obs["Status"]
        cmap_color = self.color_map(error, correct)
        # cmap_color: Callable[[int, int], str] = (
        #     lambda idx, pos: self.ERROR_COLOR
        #     if error and idx == error[pos]
        #     else self.REGULAR_COLOR
        # )
        title_color: Callable[[int, int], str] = (
            lambda idx, pos: self.ERROR_TITLE_COLOR
            if error and idx == error[pos]
            else self.REGULAR_TITLE_COLOR
        )
        node_ax: Callable[[int], npt.NDArray] = lambda n_idx: self.axs[
            n_idx // self.nodes_n_columns, n_idx % self.nodes_n_columns
        ]
        job_ax: Callable[[int], npt.NDArray] = lambda j_idx: self.axs[
            j_idx // self.jobs_n_columns,
            self.nodes_n_columns + (j_idx % self.jobs_n_columns),
        ]
        # update matries
        for n_idx, node in enumerate(nodes):
            self._draw_node(
                node,
                title_color=title_color(n_idx, 0),
                idx=n_idx,
                ax=node_ax(n_idx),
                time=self.time,
                resource=self.resource,
                cmap=cmap_color(n_idx, 0),
            )
        for j_idx, job in enumerate(queue):
            self._draw_job(
                job,
                title_color=title_color(j_idx, 1),
                idx=j_idx,
                ax=job_ax(j_idx),
                time=self.time,
                resource=self.resource,
                cmap=cmap_color(j_idx, 1),
                status=JobStatus(status[j_idx]),
            )
        # update figure
        plt.draw()
        plt.pause(self.cooldown)
        return self.fig
